
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "mac.h"

#define EPS	1e-3	/* difference of X-ray energies in keV to judge AE */

typedef struct {
		double	keV,MAC;
	} keV_MAC;

#include "log_mac.h"

static struct {
		double	mass,weight;
		size_t	size;
		keV_MAC	*p;
	} AtomicTable[]={{ 1.0/0.99212,0.0,sizeof _H__/sizeof *_H__,_H__},
			 { 2.0/0.49968,0.0,sizeof _He_/sizeof *_He_,_He_},
			 { 3.0/0.43221,0.0,sizeof _Li_/sizeof *_Li_,_Li_},
			 { 4.0/0.44384,0.0,sizeof _Be_/sizeof *_Be_,_Be_},
			 { 5.0/0.46245,0.0,sizeof _B__/sizeof *_B__,_B__},
			 { 6.0/0.49954,0.0,sizeof _C__/sizeof *_C__,_C__},
			 { 7.0/0.49976,0.0,sizeof _N__/sizeof *_N__,_N__},
			 { 8.0/0.50002,0.0,sizeof _O__/sizeof *_O__,_O__},
			 { 9.0/0.47372,0.0,sizeof _F__/sizeof *_F__,_F__},
			 {10.0/0.49555,0.0,sizeof _Ne_/sizeof *_Ne_,_Ne_},
			 {11.0/0.47847,0.0,sizeof _Na_/sizeof *_Na_,_Na_},
			 {12.0/0.49373,0.0,sizeof _Mg_/sizeof *_Mg_,_Mg_},
			 {13.0/0.48181,0.0,sizeof _Al_/sizeof *_Al_,_Al_},
			 {14.0/0.49848,0.0,sizeof _Si_/sizeof *_Si_,_Si_},
			 {15.0/0.48428,0.0,sizeof _P__/sizeof *_P__,_P__},
			 {16.0/0.49897,0.0,sizeof _S__/sizeof *_S__,_S__},
			 {17.0/0.47951,0.0,sizeof _Cl_/sizeof *_Cl_,_Cl_},
			 {18.0/0.45059,0.0,sizeof _Ar_/sizeof *_Ar_,_Ar_},
			 {19.0/0.48595,0.0,sizeof _K__/sizeof *_K__,_K__},
			 {20.0/0.49903,0.0,sizeof _Ca_/sizeof *_Ca_,_Ca_},
			 {21.0/0.46712,0.0,sizeof _Sc_/sizeof *_Sc_,_Sc_},
			 {22.0/0.45948,0.0,sizeof _Ti_/sizeof *_Ti_,_Ti_},
			 {23.0/0.45150,0.0,sizeof _V__/sizeof *_V__,_V__},
			 {24.0/0.46157,0.0,sizeof _Cr_/sizeof *_Cr_,_Cr_},
			 {25.0/0.45506,0.0,sizeof _Mn_/sizeof *_Mn_,_Mn_},
			 {26.0/0.46556,0.0,sizeof _Fe_/sizeof *_Fe_,_Fe_},
			 {27.0/0.45815,0.0,sizeof _Co_/sizeof *_Co_,_Co_},
			 {28.0/0.47708,0.0,sizeof _Ni_/sizeof *_Ni_,_Ni_},
			 {29.0/0.45636,0.0,sizeof _Cu_/sizeof *_Cu_,_Cu_},
			 {30.0/0.45879,0.0,sizeof _Zn_/sizeof *_Zn_,_Zn_},
			 {31.0/0.44462,0.0,sizeof _Ga_/sizeof *_Ga_,_Ga_},
			 {32.0/0.44071,0.0,sizeof _Ge_/sizeof *_Ge_,_Ge_},
			 {33.0/0.44046,0.0,sizeof _As_/sizeof *_As_,_As_},
			 {34.0/0.43060,0.0,sizeof _Se_/sizeof *_Se_,_Se_},
			 {35.0/0.43803,0.0,sizeof _Br_/sizeof *_Br_,_Br_},
			 {36.0/0.42959,0.0,sizeof _Kr_/sizeof *_Kr_,_Kr_},
			 {37.0/0.43291,0.0,sizeof _Rb_/sizeof *_Rb_,_Rb_},
			 {38.0/0.43369,0.0,sizeof _Sr_/sizeof *_Sr_,_Sr_},
			 {39.0/0.43867,0.0,sizeof _Y__/sizeof *_Y__,_Y__},
			 {40.0/0.43848,0.0,sizeof _Zr_/sizeof *_Zr_,_Zr_},
			 {41.0/0.44130,0.0,sizeof _Nb_/sizeof *_Nb_,_Nb_},
			 {42.0/0.43777,0.0,sizeof _Mo_/sizeof *_Mo_,_Mo_},
			 {43.0/0.43919,0.0,sizeof _Tc_/sizeof *_Tc_,_Tc_},
			 {44.0/0.43534,0.0,sizeof _Ru_/sizeof *_Ru_,_Ru_},
			 {45.0/0.43729,0.0,sizeof _Rh_/sizeof *_Rh_,_Rh_},
			 {46.0/0.43225,0.0,sizeof _Pd_/sizeof *_Pd_,_Pd_},
			 {47.0/0.43572,0.0,sizeof _Ag_/sizeof *_Ag_,_Ag_},
			 {48.0/0.42700,0.0,sizeof _Cd_/sizeof *_Cd_,_Cd_},
			 {49.0/0.42676,0.0,sizeof _In_/sizeof *_In_,_In_},
			 {50.0/0.42120,0.0,sizeof _Sn_/sizeof *_Sn_,_Sn_},
			 {51.0/0.41889,0.0,sizeof _Sb_/sizeof *_Sb_,_Sb_},
			 {52.0/0.40752,0.0,sizeof _Te_/sizeof *_Te_,_Te_},
			 {53.0/0.41764,0.0,sizeof _I__/sizeof *_I__,_I__},
			 {54.0/0.41130,0.0,sizeof _Xe_/sizeof *_Xe_,_Xe_},
			 {55.0/0.41383,0.0,sizeof _Cs_/sizeof *_Cs_,_Cs_},
			 {56.0/0.40779,0.0,sizeof _Ba_/sizeof *_Ba_,_Ba_},
			 {57.0/0.41035,0.0,sizeof _La_/sizeof *_La_,_La_},
			 {58.0/0.41395,0.0,sizeof _Ce_/sizeof *_Ce_,_Ce_},
			 {59.0/0.41871,0.0,sizeof _Pr_/sizeof *_Pr_,_Pr_},
			 {60.0/0.41597,0.0,sizeof _Nd_/sizeof *_Nd_,_Nd_},
			 {61.0/0.42094,0.0,sizeof _Pm_/sizeof *_Pm_,_Pm_},
			 {62.0/0.41234,0.0,sizeof _Sm_/sizeof *_Sm_,_Sm_},
			 {63.0/0.41457,0.0,sizeof _Eu_/sizeof *_Eu_,_Eu_},
			 {64.0/0.40699,0.0,sizeof _Gd_/sizeof *_Gd_,_Gd_},
			 {65.0/0.40900,0.0,sizeof _Tb_/sizeof *_Tb_,_Tb_},
			 {66.0/0.40615,0.0,sizeof _Dy_/sizeof *_Dy_,_Dy_},
			 {67.0/0.40623,0.0,sizeof _Ho_/sizeof *_Ho_,_Ho_},
			 {68.0/0.40655,0.0,sizeof _Er_/sizeof *_Er_,_Er_},
			 {69.0/0.40844,0.0,sizeof _Tm_/sizeof *_Tm_,_Tm_},
			 {70.0/0.40453,0.0,sizeof _Yb_/sizeof *_Yb_,_Yb_},
			 {71.0/0.40579,0.0,sizeof _Lu_/sizeof *_Lu_,_Lu_},
			 {72.0/0.40338,0.0,sizeof _Hf_/sizeof *_Hf_,_Hf_},
			 {73.0/0.40343,0.0,sizeof _Ta_/sizeof *_Ta_,_Ta_},
			 {74.0/0.40250,0.0,sizeof _W__/sizeof *_W__,_W__},
			 {75.0/0.40278,0.0,sizeof _Re_/sizeof *_Re_,_Re_},
			 {76.0/0.39958,0.0,sizeof _Os_/sizeof *_Os_,_Os_},
			 {77.0/0.40058,0.0,sizeof _Ir_/sizeof *_Ir_,_Ir_},
			 {78.0/0.39984,0.0,sizeof _Pt_/sizeof *_Pt_,_Pt_},
			 {79.0/0.40108,0.0,sizeof _Au_/sizeof *_Au_,_Au_},
			 {80.0/0.39882,0.0,sizeof _Hg_/sizeof *_Hg_,_Hg_},
			 {81.0/0.39631,0.0,sizeof _Tl_/sizeof *_Tl_,_Tl_},
			 {82.0/0.39575,0.0,sizeof _Pb_/sizeof *_Pb_,_Pb_},
			 {83.0/0.39717,0.0,sizeof _Bi_/sizeof *_Bi_,_Bi_},
			 {84.0/0.40195,0.0,sizeof _Po_/sizeof *_Po_,_Po_},
			 {85.0/0.40479,0.0,sizeof _At_/sizeof *_At_,_At_},
			 {86.0/0.38736,0.0,sizeof _Rn_/sizeof *_Rn_,_Rn_},
			 {87.0/0.39010,0.0,sizeof _Fr_/sizeof *_Fr_,_Fr_},
			 {88.0/0.38934,0.0,sizeof _Ra_/sizeof *_Ra_,_Ra_},
			 {89.0/0.39202,0.0,sizeof _Ac_/sizeof *_Ac_,_Ac_},
			 {90.0/0.38787,0.0,sizeof _Th_/sizeof *_Th_,_Th_},
			 {91.0/0.39388,0.0,sizeof _Pa_/sizeof *_Pa_,_Pa_},
			 {92.0/0.38651,0.0,sizeof _U__/sizeof *_U__,_U__}};

static int	AtomicNumber(char *element)
{
	switch(element[0]+element[1]*256) {
	    case 'H': return 1;
	    case 'H'+'e'*256: return 2;
	    case 'L'+'i'*256: return 3;
	    case 'B'+'e'*256: return 4;
	    case 'B': return 5;
	    case 'C': return 6;
	    case 'N': return 7;
	    case 'O': return 8;
	    case 'F': return 9;
	    case 'N'+'e'*256: return 10;
	    case 'N'+'a'*256: return 11;
	    case 'M'+'g'*256: return 12;
	    case 'A'+'l'*256: return 13;
	    case 'S'+'i'*256: return 14;
	    case 'P': return 15;
	    case 'S': return 16;
	    case 'C'+'l'*256: return 17;
	    case 'A'+'r'*256: return 18;
	    case 'K': return 19;
	    case 'C'+'a'*256: return 20;
	    case 'S'+'c'*256: return 21;
	    case 'T'+'i'*256: return 22;
	    case 'V': return 23;
	    case 'C'+'r'*256: return 24;
	    case 'M'+'n'*256: return 25;
	    case 'F'+'e'*256: return 26;
	    case 'C'+'o'*256: return 27;
	    case 'N'+'i'*256: return 28;
	    case 'C'+'u'*256: return 29;
	    case 'Z'+'n'*256: return 30;
	    case 'G'+'a'*256: return 31;
	    case 'G'+'e'*256: return 32;
	    case 'A'+'s'*256: return 33;
	    case 'S'+'e'*256: return 34;
	    case 'B'+'r'*256: return 35;
	    case 'K'+'r'*256: return 36;
	    case 'R'+'b'*256: return 37;
	    case 'S'+'r'*256: return 38;
	    case 'Y': return 39;
	    case 'Z'+'r'*256: return 40;
	    case 'N'+'b'*256: return 41;
	    case 'M'+'o'*256: return 42;
	    case 'T'+'c'*256: return 43;
	    case 'R'+'u'*256: return 44;
	    case 'R'+'h'*256: return 45;
	    case 'P'+'d'*256: return 46;
	    case 'A'+'g'*256: return 47;
	    case 'C'+'d'*256: return 48;
	    case 'I'+'n'*256: return 49;
	    case 'S'+'n'*256: return 50;
	    case 'S'+'b'*256: return 51;
	    case 'T'+'e'*256: return 52;
	    case 'I': return 53;
	    case 'X'+'e'*256: return 54;
	    case 'C'+'s'*256: return 55;
	    case 'B'+'a'*256: return 56;
	    case 'L'+'a'*256: return 57;
	    case 'C'+'e'*256: return 58;
	    case 'P'+'r'*256: return 59;
	    case 'N'+'d'*256: return 60;
	    case 'P'+'m'*256: return 61;
	    case 'S'+'m'*256: return 62;
	    case 'E'+'u'*256: return 63;
	    case 'G'+'d'*256: return 64;
	    case 'T'+'b'*256: return 65;
	    case 'D'+'y'*256: return 66;
	    case 'H'+'o'*256: return 67;
	    case 'E'+'r'*256: return 68;
	    case 'T'+'m'*256: return 69;
	    case 'Y'+'b'*256: return 70;
	    case 'L'+'u'*256: return 71;
	    case 'H'+'f'*256: return 72;
	    case 'T'+'a'*256: return 73;
	    case 'W': return 74;
	    case 'R'+'e'*256: return 75;
	    case 'O'+'s'*256: return 76;
	    case 'I'+'r'*256: return 77;
	    case 'P'+'t'*256: return 78;
	    case 'A'+'u'*256: return 79;
	    case 'H'+'g'*256: return 80;
	    case 'T'+'l'*256: return 81;
	    case 'P'+'b'*256: return 82;
	    case 'B'+'i'*256: return 83;
	    case 'P'+'o'*256: return 84;
	    case 'A'+'t'*256: return 85;
	    case 'R'+'n'*256: return 86;
	    case 'F'+'r'*256: return 87;
	    case 'R'+'a'*256: return 88;
	    case 'A'+'c'*256: return 89;
	    case 'T'+'h'*256: return 90;
	    case 'P'+'a'*256: return 91;
	    case 'U': return 92;
	}
	return 0;
}

static int	stack[sizeof AtomicTable/sizeof *AtomicTable],count=0;
static double	weight=0.0;

#define EMD	" \"element:mol\" description.\n"

void	SetAtomicTable(char *e_m)
{
	char	e[3];
	int	i,n;
	double	m;

	while (*e_m=='\t' || *e_m==' ') ++e_m;

	if (*e_m=='\0') {
	    (void)fputs("missing" EMD,stderr); exit(1);
	}
	if ((e[0]=e_m[0])==':')
	    i=(-1); 
	else if ((e[1]=e_m[1])==':') {
	    e[1]='\0'; i=AtomicNumber(e)-1;
	    n=(e_m[2]=='\t' || e_m[2]==' ')?0:sscanf(e_m+2,"%lf",&m);
	}
	else if ((e[2]=e_m[2])==':') {
	    e[2]='\0'; i=AtomicNumber(e)-1;
	    n=(e_m[3]=='\t' || e_m[3]==' ')?0:sscanf(e_m+3,"%lf",&m);
	}
	else
	    i=(-1);

	if (i<0 || n!=1 || m<0.0) {
	    (void)fputs("bad" EMD,stderr); exit(1);
	}
	if (AtomicTable[i].weight<=0.0) stack[count++]=i;

	m*=AtomicTable[i].mass; AtomicTable[i].weight+=m; weight+=m;
}

void	ResetAtomicTable()
{
	while (count>0) AtomicTable[stack[--count]].weight=0.0;

	weight=0.0;
}

double	MAC(double keV)
{
	double	mac;
	int	c,i;
	size_t	size;
	keV_MAC	*p2,*p1;

	if (count==0) {
	    (void)fputs("no" EMD,stderr); exit(1);
	}
	mac=0.0;
	for  (c=0; c<count; c++) {
	    i=stack[c];

	    for (size=AtomicTable[i].size, p2=(p1=AtomicTable[i].p)+1;
		 size>2 && keV>=p2->keV;
		 size--, p1=p2++) ;

	    mac+=(AtomicTable[i].weight*
		  exp((p2->keV-p1->keV<EPS)?
		      (p1->MAC+p2->MAC)/2.0:
		      (p1->MAC*(p2->keV-keV)+
		       p2->MAC*(keV-p1->keV))/(p2->keV-p1->keV)));
	}
	return mac/weight;
}
